import json
from pathlib import Path


class WikiDataPropertyTable:
    instance = None

    @classmethod
    def get_instance(clss):
        if clss.instance is None:
            clss.instance = WikiDataPropertyTable()

        return clss.instance

    def __init__(self, property_table_path=str(Path(__file__).parent / "wikidata_property_table.json")):
        self.property_id_2_name_map = {}
        self.property_name_2_id_map = {}

        self.load_property_table(property_table_path)

    def load_property_table(self, property_table_path):
        with Path(property_table_path).open("r", encoding="utf-8") as f:
            self.property_id_2_name_map = json.load(f)
        self.property_name_2_id_map = {}
        for property_id, property_name in self.property_id_2_name_map.items():
            self.property_name_2_id_map[property_name] = property_id

    def property_id_2_name(self, property_id):
        """
        parse the property id to property name
        :param property_id: the id of property. eg. "P27"
        :return:
        """
        return self.property_id_2_name_map.get(property_id, None)

    def property_name_2_id(self, property_name):
        """
        the
        :param property_name:
        :return:
        """
        return self.property_name_2_id_map.get(property_name, None)



from sekg.graph.el.base import EntityLinkResult, EntryPointLinker
from sekg.graph.util.name_searcher import KGNameSearcher
from sekg.ir.models.base import DocumentSimModel
from sekg.ir.preprocessor.base import Preprocessor


class SingleEntryPointLinker(EntryPointLinker):

    def __init__(self, preprocessor: Preprocessor, doc_sim_model: DocumentSimModel, kg_name_searcher: KGNameSearcher):
        self.preprocessor = preprocessor
        self.doc_sim_model = doc_sim_model
        self.kg_name_searcher = kg_name_searcher

    def link(self, query, **config):
        result = {}

        keywords = self.preprocessor.extract_words_for_query(query=query)

        for keyword in keywords:
            valid_id_set = self.kg_name_searcher.search_by_keyword(keyword)
            doc_retrieval_result_list = self.doc_sim_model.search(query=query, top_num=1, valid_doc_id_set=valid_id_set)

            if len(doc_retrieval_result_list) > 0:
                doc_retrieval_result = doc_retrieval_result_list[0]

                result[keyword] = EntityLinkResult(mention=keyword, node_id=doc_retrieval_result.doc_id,
                                                   node_name=doc_retrieval_result.doc_name,
                                                   score=doc_retrieval_result.score)

        return result



import re

from nltk import WordNetLemmatizer
from nltk.corpus import stopwords

from sekg.text.extractor.domain_entity.nlp_util import SpacyNLPFactory
from sekg.text.extractor.domain_entity.relation_detection import RelationType
from sekg.text.extractor.domain_entity.word_util import WordUtil
from sekg.util.code import CodeElementNameUtil


class IdentifierInfoExtractor:
    """
        IdentifierInfoExtractor, extract use knowledge from some Identifier name,
        eg. class name, method name, variable name
    """

    # SPECIAL_VERB = {"get", "set", "parse", "push", "pop", "peak", "load", "flush", "init", "goto", "show", "finish",
    #                 "invoke",
    #                 "update", "open", "close", "import", "move", "post", "process", "lookup", "select", "clear",
    #                 "reload", "prepend", "print", "claim", "visit", "resize", "build", "convert"}
    IDENTIFIER_ENTITY = "##identifier##"  ## standing for the extractor itself

    def __init__(self):

        self.nlp = SpacyNLPFactory.create_spacy_nlp_for_identifier_extractor()

        self.stopwords = stopwords.words('english')
        self.stopwords.append("-PRON-")
        self.stopwords = set(self.stopwords)
        self.lemmatizer = WordNetLemmatizer()

        self.code_patterns = [
            re.compile(r'^(?P<ELE>[a-zA-Z0-9_]*[a-z0-9][A-Z][a-z]+[a-zA-Z0-9_]*)(<.*>)?$'),
            re.compile(r'^(?P<ELE>[a-zA-Z0-9_\.<>]+)\([a-zA-Z0-9_\,.<>)]*?$'),
            re.compile(r'^(?P<ELE>[a-zA-Z]{2,}(\.[a-zA-Z0-9_]+)+)(<.*>)?$')
        ]

        self.camel_cache = {}
        self.CODE_NAME_UTIL = CodeElementNameUtil()

    def uncamelize(self, camel_case):
        if camel_case in self.camel_cache:
            return self.camel_cache[camel_case]
        sub = self.CODE_NAME_UTIL.uncamelize_by_stemming(camel_case)
        self.camel_cache[camel_case] = sub
        return sub

    def pos_tag_for_method_name(self, name):
        """
        do the pos tag for method name
        :param name: name
        :return: a Spacy Doc object
        """
        name = name.split("(")[0].split(".")[-1]
        uncamel = self.uncamelize(name).lower()
        spacy_doc = self.nlp(uncamel)

        # fix the problem of 2 and 4 to represent 'to' and 'for'
        for i in range(0, len(spacy_doc)):
            if spacy_doc[i].text == "2" and i != 0 and len(spacy_doc) - 1:
                spacy_doc[i].lemma_ = "to"
                spacy_doc[i].pos_ = "ADP"
                spacy_doc[i].tag_ = "IN"

            if spacy_doc[i].text == "4" and i != 0 and len(spacy_doc) - 1:
                spacy_doc[i].lemma_ = "for"
                spacy_doc[i].pos_ = "ADP"
                spacy_doc[i].tag_ = "IN"

            if spacy_doc[i].text.lower() == "to":
                spacy_doc[i].lemma_ = "to"
                spacy_doc[i].pos_ = "ADP"
                spacy_doc[i].tag_ = "IN"

            if spacy_doc[i].text.lower() in ["id", "ID", "Id"]:
                spacy_doc[i].lemma_ = "id"
                spacy_doc[i].pos_ = "NOUN"
                spacy_doc[i].tag_ = "NN"

        parts = []
        current_part = []
        for token in spacy_doc:
            if token.pos_ == "ADP":
                if current_part:
                    parts.append(current_part)
                current_part = []
                parts.append([token])
            else:
                current_part.append(token)
        if current_part:
            parts.append(current_part)

        if len(parts) == 3 and parts[1][0].text == "2":
            self.__reannotation_for_tokens(parts[0], fix_to_noun_ending=True)
            self.__reannotation_for_tokens(parts[2], fix_to_noun_ending=True)
            return spacy_doc

        leading_tokens = parts[0]
        self.__reannotate_for_leading_part_tokens(leading_tokens)

        left_parts = parts[1:]
        if len(left_parts) == 0:
            return spacy_doc

        for tokens in left_parts:
            # 介词后面的作为独立的一部分，让spacy重新标注
            if len(tokens) == 1 and tokens[0].pos_ == "ADP":
                continue

            self.__reannotate_for_left_part_tokens_after_preposition(tokens)
        return spacy_doc

    def __reannotate_for_leading_part_tokens(self, part):
        text = " ".join([token.text for token in part])
        # for "find or create" type, just use the default annotation, it is OK
        if "and" in text.lower() or "or" in text.lower():
            self.__reannotation_for_tokens(part)
            return
            # todo
        if len(part) == 1:
            # first part is a ADP(eg. To,By)
            if part[0].pos_ == "ADP":
                return
            # new_token = self.nlp(part[0].text)[0]
            # self.__update_annotation_for_token(part[0],new_token)
            if WordUtil.couldBeVerb(part[0].text):
                self.__set_token_to_verb_pos(part[0])
                return
            if WordUtil.couldBeNoun(part[0].text):
                self.__set_token_to_noun_pos(part[0])
                return
            if WordUtil.couldBeADJ(part[0].text):
                self.__set_token_to_adj_pos(part[0])
                return
            return

        if len(part) >= 1:
            ## 1. VERB + NOUN
            ## 2. VERB + NOUN +ADJ
            ## 3. VERB + ADJ
            ## 4. NOUN

            left_tokens = part
            if WordUtil.couldBeVerb(part[0].text):
                self.__set_token_to_verb_pos(part[0])
                left_tokens = part[1:]

            if WordUtil.couldBeNoun(left_tokens[-1].text):
                ## 1. VERB + NOUN
                ## 4. NOUN
                self.__reannotation_for_tokens(left_tokens, fix_to_noun_ending=True)
                return

            if WordUtil.couldBeADJ(left_tokens[-1].text):
                ## 2. VERB + NOUN +ADJ
                ## 3. VERB + ADJ
                self.__set_token_to_adj_pos(left_tokens[-1])
                left_tokens = left_tokens[:-1]
                self.__reannotation_for_tokens(left_tokens, fix_to_noun_ending=True)
                return

            self.__reannotation_for_tokens(left_tokens)

    def __reannotate_for_left_part_tokens_after_preposition(self, tokens):
        text = " ".join([token.text for token in tokens])
        # for "find or create" type, just use the default annotation, it is OK
        if "and" in text.lower() or "or" in text.lower():
            self.__reannotation_for_tokens(tokens)
            return
        if len(tokens) == 1:
            # ADP+NOUN
            if WordUtil.couldBeNoun(tokens[0].text):
                self.__set_token_to_noun_pos(tokens[0])
                return
            # ADP+ADJ
            if WordUtil.couldBeADJ(tokens[0].text):
                self.__set_token_to_adj_pos(tokens[0])
                return
            # others: annotate by default
            self.__reannotation_for_tokens(tokens)
            return

        if len(tokens) >= 1:
            ## 1. NOUN
            ## 2. NOUN +ADJ
            ## 3. ADJ
            ## 4. NOUN

            if WordUtil.couldBeNoun(tokens[-1].text):
                ## 1. NOUN
                ## 4. NOUN
                self.__reannotation_for_tokens(tokens, fix_to_noun_ending=True)
                return

            if WordUtil.couldBeADJ(tokens[-1].text):
                ## 2. VERB + NOUN +ADJ
                ## 3. VERB + ADJ
                self.__set_token_to_adj_pos(tokens[-1])

                self.__reannotation_for_tokens(tokens[:-1], fix_to_noun_ending=True)
                return

            self.__reannotation_for_tokens(tokens)

    def __reannotation_for_tokens(self, tokens, fix_to_noun_ending=False):
        if len(tokens) == 0:
            return
        text = " ".join([token.text for token in tokens])
        new_tokens = self.nlp(text)
        if len(new_tokens) != len(tokens):
            for token in tokens:
                new_token = self.nlp(token.text)[0]
                self.__update_annotation_for_token(token, new_token)
        else:
            for old_annotation_token, new_annotation_token in zip(tokens, new_tokens):
                self.__update_annotation_for_token(old_annotation_token, new_annotation_token)

        self.__fix_adj_like_verb_annotation(tokens)
        self.__fix_special_vocab_annotation(tokens)
        if fix_to_noun_ending and WordUtil.couldBeNoun(tokens[-1].text):
            self.__set_token_to_noun_pos(tokens[-1])

    def __fix_adj_like_verb_annotation(self, tokens):
        for token in tokens:
            if token.tag_ in ["VBD", "VBG", "VBN"]:
                self.__set_token_to_adj_pos(token)

    def __fix_special_vocab_annotation(self, tokens):
        for token in tokens:
            if token.text.lower() == "id":
                token.lemma_ = "id"
                token.pos_ = "NOUN"
                token.tag_ = "NN"

    def __update_annotation_for_token(self, old_annotation_token, new_annotation_token):
        old_annotation_token.lemma_ = new_annotation_token.lemma_
        old_annotation_token.pos_ = new_annotation_token.pos_
        old_annotation_token.tag_ = new_annotation_token.tag_

    def __set_token_to_verb_pos(self, old_annotation_token):
        # old_annotation_token.lemma_ = new_annotation_token.lemma_
        old_annotation_token.pos_ = "VERB"
        old_annotation_token.tag_ = "VB"

    def __set_token_to_noun_pos(self, old_annotation_token):
        # old_annotation_token.lemma_ = new_annotation_token.lemma_
        old_annotation_token.pos_ = "NOUN"
        old_annotation_token.tag_ = "NN"

    def __set_token_to_adj_pos(self, old_annotation_token):
        # old_annotation_token.lemma_ = new_annotation_token.lemma_
        old_annotation_token.pos_ = "ADJ"
        old_annotation_token.tag_ = "JJ"

    def __find_depending_verbs(self, np_tokens, spacy_doc):
        spacy_doc = [token for token in spacy_doc]
        result = []
        start_token = np_tokens[0]
        start_index = spacy_doc.index(start_token)

        for i in range(start_index - 1, -1, -1):
            if spacy_doc[i].pos_ == "VERB":
                result.append(spacy_doc[i])
                if i - 2 < 0:
                    return result

                if spacy_doc[i - 1].text.lower() != "and" and spacy_doc[i - 1].text.lower() != "or":
                    return result

                if spacy_doc[i - 2].pos_ != "VERB":
                    return result
                result.append(spacy_doc[i - 2])
                return result

        return result

    def __find_adp_token_between_verb_and_np(self, spacy_doc, np_tokens, operation_token):
        spacy_doc = [token for token in spacy_doc]

        operation_token_index = spacy_doc.index(operation_token)
        start_token = np_tokens[0]
        start_index = spacy_doc.index(start_token)

        for i in range(start_index - 1, operation_token_index, -1):
            if spacy_doc[i].pos_ == "ADP":
                return spacy_doc[i]
        return None

    def __reextract_np_from_single_part(self, tokens):
        if len(tokens) == 0:
            return []

        text = " ".join([token.text for token in tokens]).lower()

        text = " ".join([token.text + "" + token.pos_ for token in tokens]).lower()
        total_nps = []
        current_np = []

        for index, token in enumerate(tokens):
            if token.pos_ == "ADP" and token.text.lower() == "of":
                if len(current_np) == 0:
                    current_np = []
                    continue

                if current_np[-1].pos_ not in ["PROPN", "NOUN"]:
                    current_np = []
                    continue

                current_np.append(token)
                continue
            if token.tag_ in ["WP$", "WDT", "WP"] or token.pos_ not in ["CCONJ", "ADJ", "PROPN", "NOUN"]:
                if len(current_np) == 0:
                    continue
                if current_np[-1].pos_ not in ["PROPN", "NOUN"]:
                    current_np = []
                    continue

                total_nps.append(current_np)
                current_np = []
                continue
            current_np.append(token)
            continue

        if len(current_np) != 0 and current_np[-1].pos_ in ["PROPN", "NOUN"]:
            total_nps.append(current_np)

        if len(total_nps) == 0:
            return []
        if len(total_nps) == 1:
            return total_nps

        return total_nps

    def __reextract_np_from_spacy_pos_doc(self, tokens):
        if len(tokens) == 0:
            return []

        parts = []
        current_part = []
        for token in tokens:
            if token.pos_ == "ADP" and token.text.lower() != "of":
                if current_part:
                    parts.append(current_part)
                current_part = []
            else:
                current_part.append(token)
        if current_part:
            parts.append(current_part)
        noun_phase_list = []

        for part in parts:

            nps = self.__reextract_np_from_single_part(part)
            if len(nps) > 0:
                noun_phase_list.extend(nps)
        return noun_phase_list

    def extract_from_method_name(self, method_name, mark_for_identifier_in_relation=IDENTIFIER_ENTITY):
        terms = set()
        operations = set()

        spacy_doc = self.pos_tag_for_method_name(method_name)
        noun_chunks = self.__reextract_np_from_spacy_pos_doc(spacy_doc)

        operations.update([token.text for token in spacy_doc if token.pos_ == "VERB"])
        terms.update([" ".join([t.text for t in np_tokens]).lower() for np_tokens in noun_chunks])

        text = " ".join([token.text for token in spacy_doc]).lower()

        # group the operation by level, eg. "create and find", "create" and "find" are the same level.
        terms = self.filter_terms(terms)

        relations = set([])
        identifier_relations = set([])
        for np_tokens in noun_chunks:
            np = " ".join([t.text for t in np_tokens]).lower()
            operation_tokens = self.__find_depending_verbs(np_tokens, spacy_doc)

            for operation_token in operation_tokens:
                relations.add((np, RelationType.CAN_BE_OPERATED.value, operation_token.lemma_.lower()))

            for operation_token in operation_tokens:
                adp_token = self.__find_adp_token_between_verb_and_np(spacy_doc, np_tokens, operation_token)
                adp_text = ""
                if adp_token == None:
                    adp_text = ""
                else:
                    adp_text = adp_token.text

                relations.add((np, RelationType.CAN_BE_OPERATED.value, operation_token.lemma_.lower()))

                operation_relation_name = ("operation_" + operation_token.lemma_.lower() + " " + adp_text).strip()
                identifier_relations.add((mark_for_identifier_in_relation, operation_relation_name, np))

        for op in operations:
            identifier_relations.add((mark_for_identifier_in_relation, RelationType.INSTANCE_OF.value, op))

        return terms, operations, relations, identifier_relations

    def extract_from_variable(self, variable, mark_for_identifier_in_relation=IDENTIFIER_ENTITY):
        """
        extract from variable name , eg. "String filePath"
        :param variable: s string that represent variable , eg. "String filePath", "java.lang.String filePath"
        :param mark_for_identifier_in_relation, the mark for the identifier itself in relations
        :return:
        """
        if not variable:
            return set(), set(), set()

        terms = set()
        variable = variable.split("<")[0].split(".")[-1].split(" ")[-1]
        if not variable:
            return set(), set(), set()
        uncamel = self.uncamelize(variable)
        for np in self.nlp(uncamel).noun_chunks:
            lemma = " ".join([token.text for token in np if token.lemma_ != "-PRON-"])
            if len(lemma) <= 1 or lemma.isdigit() == True:
                continue

            terms.add(lemma)

        terms = self.filter_terms(terms)

        relations = set()
        identifier_relations = set()

        if len(terms) == 1:
            terms_ = list(terms)[0]
            if terms_.lower() == uncamel.lower():
                identifier_relations.add((mark_for_identifier_in_relation, RelationType.REPRESENT.value, terms_))


        else:
            for lemma in terms:
                identifier_relations.add((mark_for_identifier_in_relation, RelationType.NAME_MENTION.value, lemma))
        return terms, relations, identifier_relations

    def extract_from_class_name(self, name, mark_for_identifier_in_relation=IDENTIFIER_ENTITY):
        """
        extract from class name or the name like class. eg. package name. example: java.lang.ArrayList-> Array List
        :param name:
        :param mark_for_identifier_in_relation, the mark for the identifier itself in relations

        :return:
        """
        if not name:
            return set(), set(), set()
        # todo: Error raise: IndexError: list index out of range
        name = name.split("<")[0].split(".")[-1].split(" ")[-1]
        if not name:
            return set(), set(), set()
        uncamel = self.uncamelize(name)

        terms = {uncamel}
        terms = self.filter_terms(terms)
        relations = set()
        identifier_relations = set()
        for lemma in terms:
            identifier_relations.add((mark_for_identifier_in_relation, RelationType.REPRESENT.value, lemma))
        return terms, relations, identifier_relations

    def __valid(self, term):
        term = str(term).lower()
        if len(term) <= 1 or term.isdigit():
            return False
        if term in self.stopwords:
            return False
        if len(set(term)) == 1:
            return False

        return True

    def filter_terms(self, terms):
        valid_terms = set()
        for t in terms:
            if self.__valid(t):
                valid_terms.add(t)

        return valid_terms

    # following is the version of wc to implement to extract relation from method name
    # def extract_from_method_name_wc_version(self, name):
    #     terms = set()
    #     operations = set()
    #     name = name.split("(")[0].split(".")[-1]
    #     uncamel = self.uncamelize(name).lower()
    #     words = uncamel.split()
    #     print("words", words)
    #
    #     spacy_doc = self.nlp(uncamel)
    #
    #     pos_tags = [token.pos_ for token in spacy_doc]
    #     lemma_words = [token.lemma_ for token in spacy_doc]
    #     np_index = {np.start: np.end for np in spacy_doc.noun_chunks}
    #     print("before fix the tags and lemma", pos_tags, lemma_words)
    #
    #     if len(words) == 1:
    #         if pos_tags[0] == "VERB":
    #             return set(), {words[0]}
    #         if pos_tags[0] == "NOUN":
    #             return {words[0]}, set()
    #
    #         return terms, {words[0]}
    #
    #     parts = []
    #     index = 0
    #     word_num = len(lemma_words)
    #     adj_flag = -1
    #     while index < word_num:
    #         if index in np_index:
    #             np_words = lemma_words[index: np_index[index]]
    #             if np_words[0] in self.SPECIAL_VERB:
    #                 parts.append((np_words[0], "VERB"))
    #                 np = " ".join(np_words[1:])
    #             else:
    #                 np = " ".join(np_words)
    #             if len(np) > 0:
    #                 if adj_flag >= 0:
    #                     np = lemma_words[adj_flag] + " " + np
    #                 parts.append((np, "NP"))
    #             index = np_index[index]
    #             adj_flag = -1
    #         elif lemma_words[index] in {"id", "ID", "Id"}:
    #             np = lemma_words[index]
    #             if adj_flag >= 0:
    #                 np = lemma_words[adj_flag] + " " + np
    #
    #             elif len(parts) > 0 and parts[-1][1] == "NP":
    #                 ele, tag = parts.pop()
    #                 np = "{} {}".format(ele, lemma_words[index])
    #             parts.append((np, "NP"))
    #             index += 1
    #             adj_flag = -1
    #         elif pos_tags[index] == "VERB" and lemma_words[index].endswith("ed"):
    #             adj_flag = index
    #             index += 1
    #         else:
    #             adj_flag = -1
    #             parts.append((lemma_words[index], pos_tags[index]))
    #             index += 1
    #     index = 0
    #     np_flag = -1
    #     of_flag = -1
    #     while index < len(parts):
    #         ele, tag = parts[index]
    #         if tag == "NP":
    #             if of_flag >= 0:
    #                 cur = parts.pop(index)[0]
    #                 of = parts.pop(index - 1)[0]
    #                 prev = parts.pop(index - 2)[0]
    #                 np = " ".join([prev, of, cur])
    #                 parts.append((np, "NP"))
    #                 index = index - 2
    #                 of_flag = -1
    #             else:
    #                 np_flag = index
    #                 index += 1
    #         elif ele == "of" and of_flag < 0 and np_flag >= 0:
    #             of_flag = index
    #             index += 1
    #         else:
    #             np_flag = -1
    #             of_flag = -1
    #             index += 1
    #     # print(parts)
    #     operations.update([ele for ele, tag in parts if tag == "VERB"])
    #     nps = {ele for ele, tag in parts if tag == "NP"}
    #     terms.update(nps)
    #     print("pos tags", parts)
    #
    #     return terms, operations



from unittest import TestCase

from sekg.wiki.search_domain_wiki.search import WikiTool


class TestWikiTool(TestCase):

    def test_wiki_tool(self):
        test_set = set()
        test_set.add("Peer alarm")
        wiki_tool = WikiTool(domain_name_set_to_search=test_set, use_proxies=False, max_candidate_num=5)
        # wiki_tool.start_search()
        # p = "."
        # wiki_tool.save(p)
        # w = WikiTool(domain_name_set_to_search=set(), use_proxies=False, max_candidate_num=5)
        # w.load(p)
        # print(w.title_2_wiki)
        result = wiki_tool.search_title("Peer alarm")
        print(result)



import math
import networkx as nx
from gensim.models.keyedvectors import Word2VecKeyedVectors
from nltk import WordNetLemmatizer

from sekg.constant.code import CodeEntityCategory
from sekg.constant.constant import OperationConstance, DomainConstant
from sekg.graph.el.base import EntryPointLinker, EntityLinkResult
from sekg.graph.exporter.graph_data import GraphData
from sekg.graph.util.name_searcher import KGNameSearcher
from sekg.ir.models.base import DocumentSimModel
from sekg.ir.preprocessor.base import Preprocessor


class BestSeqSelector:

    def __init__(self, ):
        self.keyword2candidate_map = {}
        self.G = nx.Graph()
        self.best_selection = []
        self.best_score = 0.0

    def get_candidate_ids(self):
        result = set([])
        for keyword, ids in self.keyword2candidate_map.items():
            result = result | ids
        return list(result)

    def get_keyword_list(self):
        return list(self.keyword2candidate_map.keys())

    def add_candidate_for_keyword(self, keyword, candidate_node_id, score):
        if keyword not in self.keyword2candidate_map:
            self.keyword2candidate_map[keyword] = set([])

        self.keyword2candidate_map[keyword].add(candidate_node_id)
        self.add_candidate_entity_score(node_id=candidate_node_id, score=score)

    def add_candidate_entity_score(self, node_id, score):
        self.G.add_node(node_id, score=score)

    def add_pair_score(self, start_id, end_id, score):
        self.G.add_edge(start_id, end_id, score=score)

    def search_best_combination(self):
        self.chosen([], score=0.0, current_chosen_keyword_index=0)

        return self.get_best_selection()

    def get_best_score(self):
        return self.best_score

    def get_best_selection(self):
        keyword_list = self.get_keyword_list()
        result = {}
        for keyword, node_id in zip(keyword_list, self.best_selection):
            result[keyword] = node_id
        return result

    def chosen(self, history_chosen, score, current_chosen_keyword_index):
        if current_chosen_keyword_index >= len(self.get_keyword_list()):
            if self.best_score < score:
                self.best_score = score
                self.best_selection = history_chosen
            else:
                return
        else:
            current_keyword = self.get_keyword_list()[current_chosen_keyword_index]

            candidate_ids = self.keyword2candidate_map[current_keyword]

            for candidate_entity_id in candidate_ids:
                extra_score = self.G.nodes[candidate_entity_id]["score"]

                for history_chosen_entity_id in history_chosen:
                    extra_score += self.G[candidate_entity_id][history_chosen_entity_id]["score"]

                self.chosen(history_chosen + [candidate_entity_id], score=score + extra_score,
                            current_chosen_keyword_index=current_chosen_keyword_index + 1)


class NFGlobalEntryPointLinker(EntryPointLinker):
    DEFAULT_CANDIDATE_NUM_FOR_EVERY_POINT = 3

    def __init__(self, preprocessor: Preprocessor,
                 doc_sim_model: DocumentSimModel,
                 kg_name_searcher: KGNameSearcher,
                 graph2vecModel: Word2VecKeyedVectors,
                 graph_data: GraphData,
                 ):
        self.preprocessor = preprocessor
        self.doc_sim_model = doc_sim_model
        self.graph2vecModel = graph2vecModel
        self.kg_name_searcher = kg_name_searcher
        self.graph_data = graph_data
        self.lemmatizer = WordNetLemmatizer()
        # to start the util
        self.lemmatizer.lemmatize(word="test")

    def score_full_name(self, candidate_node_ids, keywords, name_score_map, original_query):
        new_keywords = set([])
        for keyword in keywords:
            new_keywords.add(keyword.lower())
        keywords = new_keywords
        clean_query_string = " ".join(list(keywords))

        for node_id in candidate_node_ids:
            if node_id not in name_score_map:
                name_score_map[node_id] = self.compute_name_score_for_one_node(keywords, clean_query_string, node_id,
                                                                               original_query)
        return name_score_map

    def compute_name_score_for_one_node(self, keywords, query_string, node_id, original_query):
        full_names = self.kg_name_searcher.get_full_names(node_id)
        score_list = []
        for full_name in full_names:
            full_name = full_name.lower()
            name_set = set(full_name.split(" "))
            word_count = len(name_set)
            token_score = 1.0 * len(name_set & keywords) / word_count

            appearing = 0.0
            for word in name_set:
                if word in query_string or word in original_query:
                    appearing += 1

            appearing_score = appearing / word_count
            score_list.append(appearing_score)

            score_list.append(token_score)

        if score_list:
            return max(score_list)
        return 0.0

    def select_candidates_for_keyword(self, query, keywords, candidate_num):

        keyword_to_candidate_list_map = {}
        name_score_map = {}

        for keyword in keywords:
            doc_retrieval_result_list = self.get_candidate_for_keyword(candidate_num, keyword, query, name_score_map,
                                                                       keywords)
            if len(doc_retrieval_result_list) > 0:
                keyword_to_candidate_list_map[keyword] = doc_retrieval_result_list

        return keyword_to_candidate_list_map


    def link(self, query, candidate_num=DEFAULT_CANDIDATE_NUM_FOR_EVERY_POINT):
        selector = BestSeqSelector()
        keywords = self.preprocessor.extract_words_for_query(query=query)

        doc_id_2_doc_retrieval_result_map = {}

        keyword_to_candidate_list_map = self.select_candidates_for_keyword(query=query, keywords=keywords,
                                                                           candidate_num=candidate_num)

        for keyword, doc_retrieval_result_list in keyword_to_candidate_list_map.items():

            for doc_retrieval_result in doc_retrieval_result_list:
                # print("candidate keyword=%s %r" % (keyword, doc_retrieval_result))
                selector.add_candidate_for_keyword(keyword=keyword,
                                                   candidate_node_id=doc_retrieval_result.doc_id,
                                                   score=doc_retrieval_result.score)

                doc_id_2_doc_retrieval_result_map[doc_retrieval_result.doc_id] = doc_retrieval_result
        candidate_ids = selector.get_candidate_ids()

        vector_map = {}

        all_vectors = []

        for candidate_id in candidate_ids:
            vector_map[candidate_id] = self.graph2vecModel.wv[str(candidate_id)]
            all_vectors.append(vector_map[candidate_id])

        if len(keywords) == 1:
            rate = 0.0
        else:
            rate = 2.0 / (len(keywords) - 1.0)

        for start_id in candidate_ids:
            start_vector = vector_map[start_id]
            sim_vector = self.graph2vecModel.wv.cosine_similarities(start_vector, all_vectors)

            for end_id, score in zip(candidate_ids, sim_vector):
                # print("pair score =%r %r %r" % (start_id, end_id, score))
                selector.add_pair_score(start_id=start_id, end_id=end_id, score=score * rate)

        best_linking_result = selector.search_best_combination()

        result = {}
        for keyword, node_id in best_linking_result.items():
            doc_retrieval_result = doc_id_2_doc_retrieval_result_map[node_id]

            result[keyword] = EntityLinkResult(mention=keyword, node_id=node_id,
                                               node_name=doc_retrieval_result.doc_name,
                                               score=doc_retrieval_result.score)

        return result

    def get_candidate_for_keyword(self, candidate_num, keyword, query, name_score_map, keywords):

        candidate_doc_result_list = []

        full_name_match_id_set = self.kg_name_searcher.search_by_full_name(keyword.lower())
        keyword_match_id_set = self.kg_name_searcher.search_by_keyword(keyword.lower())

        new_keyword = self.lemmatizer.lemmatize(word=keyword, pos="v")
        new_keyword = new_keyword.lower()

        extra_veb_keywords = set([])
        if new_keyword != keyword:
            t = self.kg_name_searcher.search_by_full_name(new_keyword)
            full_name_match_id_set = full_name_match_id_set | t
            extra_veb_keywords = extra_veb_keywords | t
            t = self.kg_name_searcher.search_by_keyword(new_keyword)
            keyword_match_id_set = keyword_match_id_set | t
            extra_veb_keywords = extra_veb_keywords | t

        ## todo: this is only for code kg
        operation_ids = self.graph_data.get_node_ids_by_label(OperationConstance.LABEL_OPERATION)
        domain_term_ids = self.graph_data.get_node_ids_by_label(DomainConstant.LABEL_DOMAIN_TERM)

        method_ids = self.graph_data.get_node_ids_by_label(
            CodeEntityCategory.to_str(CodeEntityCategory.CATEGORY_METHOD))
        method_ids = method_ids | self.graph_data.get_node_ids_by_label(
            CodeEntityCategory.to_str(CodeEntityCategory.CATEGORY_BASE_OVERRIDE_METHOD))

        class_field_ids = self.graph_data.get_node_ids_by_label(
            CodeEntityCategory.to_str(CodeEntityCategory.CATEGORY_FIELD_OF_CLASS))

        if len(operation_ids & full_name_match_id_set) > 0 or len(domain_term_ids & full_name_match_id_set) > 0:
            full_name_match_id_set = full_name_match_id_set - method_ids

            keyword_match_id_set = keyword_match_id_set - (full_name_match_id_set & method_ids)

        if len(class_field_ids & full_name_match_id_set) > 0:
            full_name_match_id_set = full_name_match_id_set - class_field_ids

            keyword_match_id_set = keyword_match_id_set - (full_name_match_id_set & class_field_ids)

        chosen_ids = set([])

        large_candidate_num = candidate_num + 5
        if len(full_name_match_id_set) > 0:
            doc_retrieval_result_list = self.doc_sim_model.search(query=query, top_num=large_candidate_num,
                                                                  valid_doc_id_set=full_name_match_id_set)

            for doc in doc_retrieval_result_list:
                if doc.doc_id not in chosen_ids:
                    chosen_ids.add(doc.doc_id)
                    candidate_doc_result_list.append(doc)

        if len(keyword_match_id_set) > 0:
            doc_retrieval_result_list = self.doc_sim_model.search(query=query, top_num=large_candidate_num,
                                                                  valid_doc_id_set=keyword_match_id_set)

            for doc in doc_retrieval_result_list:
                if doc.doc_id not in chosen_ids:
                    chosen_ids.add(doc.doc_id)
                    candidate_doc_result_list.append(doc)

        self.score_full_name(candidate_node_ids=chosen_ids, keywords=keywords, name_score_map=name_score_map,
                             original_query=query)

        for candidate_doc in candidate_doc_result_list:
            if candidate_doc.doc_id in extra_veb_keywords:
                candidate_doc.score += 1.0 / (math.log10(len(extra_veb_keywords)) + 1)

            candidate_doc.extra_info["name_score"] = name_score_map[candidate_doc.doc_id]
            candidate_doc.extra_info["doc_score"] = candidate_doc.score
            candidate_doc.score = candidate_doc.extra_info["name_score"] + candidate_doc.extra_info["doc_score"]

        candidate_doc_result_list = sorted(candidate_doc_result_list, key=lambda candidate_doc: candidate_doc.score,
                                           reverse=True)

        # for t in candidate_doc_result_list:
        #     print("name score=%r doc_score %r %r %r" % (
        #         t.extra_info["name_score"], t.extra_info["doc_score"], t, self.graph_data.get_node_info_dict(t.doc_id)))
        # print("--------")
        if len(candidate_doc_result_list) >= 2:
            if candidate_doc_result_list[0].score / candidate_doc_result_list[1].score > 1.5:
                return [candidate_doc_result_list[0]]

        return candidate_doc_result_list[:candidate_num]



import re

import nltk
from nltk import WordNetLemmatizer
from nltk.corpus import stopwords

from sekg.util.code import CodeElementNameUtil
from .base import Preprocessor


class CodeDocPreprocessor(Preprocessor):
    """
    The new way we design for preprocess text containing code element. split camel case and add the camel word back to the text.
    Recommend to use for our KG document.
    """
    CLEAN_PATTERN = re.compile(r'[^a-zA-Z_]')
    STOPWORDS = set(stopwords.words('english'))
    code_element_util = CodeElementNameUtil()
    lemmatizer = WordNetLemmatizer()
    lemma_map = {}

    def extract_words_for_query(self, query):
        return set(self.clean(query))

    def lemma(self, word):
        if word in self.lemma_map:
            return self.lemma_map[word]

        lemma = self.lemmatizer.lemmatize(self.lemmatizer.lemmatize(word, "n"), "v")

        if lemma != word:
            self.lemma_map[word] = lemma
        return lemma

    def clean(self, text):
        """
        return a list of token from the text, only remove stopword and lemma the word
        :param text: the text need to preprocess
        :return: list of str
        """

        clean_text = re.sub(self.CLEAN_PATTERN, " ", text)
        old_words_set = set([word.lower() for word in clean_text.split(" ") if word and word not in self.STOPWORDS])
        clean_text = self.code_element_util.uncamelize_from_simple_name(clean_text)
        if not clean_text:
            return []
        new_words = clean_text.lower().split(" ")
        new_words = [self.lemma(word) for word in new_words
                     if word and word not in self.STOPWORDS]

        new_words_set = set(new_words)

        for old_word in old_words_set:
            if old_word not in new_words_set and old_word not in self.lemma_map:
                new_words.append(old_word)

        return new_words

class PureCodePreprocessor(Preprocessor):
    """
    traditional preprocessing way for feature location, split camel word, remove keyword and stop words.
    Todo: Add paper citation.
    """
    stemmer = nltk.PorterStemmer(nltk.PorterStemmer.MARTIN_EXTENSIONS)
    stops = frozenset(nltk.corpus.stopwords.words('english'))
    Java_Keyword = {"abstract", "assert", "boolean", "break", "byte", "case", "catch", "char", "class", "const",
                    "continue", "default", "do", "double", "else", "enum", "extends", "final", "finally", "float",
                    "for", "goto", "if", "implements", "import", "instanceof", "int", "interface", "long", "native",
                    "new", "package", "private", "protected", "public", "return", "strictfp", "short", "static",
                    "super", "switch", "this", "throw", "throws", "try", "void", "transient", "synchronized",
                    "volatile", "while"}
    remove_list = {"\n", "\t", "\r", "/", "*", ".", ";", "@", "{", "}", "<p>", "(", ")", "#", "=", ":", "+", "-",
                   "!",
                   "[", "]", ",", ":", "<", ">", "|", "\\", "&", "'", "?", "<", ">", "'", "`", "\""}

    def hump_split(self, hunp_str):
        p = re.compile(r'([a-z]|\d)([A-Z])')
        sub = re.sub(p, r'\1_\2', hunp_str).lower()
        if "_" not in sub:
            return [sub]
        sub = sub.split("_")
        return sub

    def stems(self, stops, stemmer, string):
        to_skip = set('()!@#$%^&*-+=|{}[]<>./?;:')
        to_skip.update(self.Java_Keyword)

        numeric_regex = re.compile(r'(^-?\d+\.\d+$)|(^-?\d+$)')

        if string in to_skip:
            raise StopIteration
        for word in nltk.tokenize.word_tokenize(string):
            if word in to_skip:
                continue

            if word in stops or '/' in word:
                continue

            if numeric_regex.match(word) is not None:
                continue

            try:
                yield stemmer.stem(word).lower()
            except Exception:
                continue

    def remove_special_char(self, data):
        """
        input;str
        :return: str
        """
        new_str = data
        for item in self.remove_list:
            new_str = new_str.replace(item, " ").replace("  ", " ")
        return new_str.strip()

    def clean(self, data):
        data = self.remove_special_char(data)
        return_list = []
        for item in data.split(" "):
            return_list = return_list + self.hump_split(item)
        obs = list(self.stems(self.stops, self.stemmer, " ".join(return_list)))
        return obs


from unittest import TestCase

from py2neo import Graph

from sekg.graph.accessor import GraphAccessor
from sekg.graph.creator import NodeBuilder
from sekg.graph.factory import GraphInstanceFactory
from sekg.graph.label_accessor import LabelGraphAccessor


class TestGraphAccessor(TestCase):
    test_server_name = "LocalHost"
    test_config_path = "neo4j_config.json"
    test_server_id = 3

    def test_get_graph(self):
        factory = GraphInstanceFactory(self.test_config_path)
        graph = factory.create_py2neo_graph_by_server_id(self.test_server_id)
        accessor = GraphAccessor(graph=graph)
        self.assertTrue(isinstance(accessor.graph, Graph))

    def test_is_connect(self):
        factory = GraphInstanceFactory(self.test_config_path)
        graph = factory.create_py2neo_graph_by_server_id(self.test_server_id)
        accessor = GraphAccessor(graph=None)
        self.assertFalse(accessor.is_connect())
        accessor = GraphAccessor(graph=graph)
        self.assertTrue(accessor.is_connect())

    def test_create_or_drop_index_node(self):
        factory = GraphInstanceFactory(self.test_config_path)
        graph = factory.create_py2neo_graph_by_server_id(self.test_server_id)

        accessor = LabelGraphAccessor(graph=graph)
        self.assertTrue(accessor.is_connect())

        node = NodeBuilder().add_entity_label().add_label("api class").add_one_property(property_name="api id",
                                                                                        property_value=4).add_one_property(
            property_name="api name", property_value="java.lang.Object").build()

        node = accessor.create_or_update_node(node, primary_label="test", primary_property="api id")
        print(node)

        accessor.add_label_for_one_label_node("api class", "java api class")
        node = accessor.find_node(primary_label="api class", primary_property="api id", primary_property_value=4)
        self.assertEqual(len(node.labels), 3)

        print(node)
        accessor.delete_label_for_one_label_node("api class", "java api class")
        node = accessor.find_node(primary_label="api class", primary_property="api id", primary_property_value=4)

        self.assertEqual(len(node.labels), 2)

        accessor.delete_label_for_one_label_node("java api class", "java api class")
        node = accessor.find_node(primary_label="api class", primary_property="api id", primary_property_value=4)
        self.assertEqual(len(node.labels), 2)

        node = accessor.find_node(primary_label="api class", primary_property="api id", primary_property_value=4)
        self.assertEqual(len(node.labels), 2)

        graph.delete(node)



import gensim

from sekg.graph.exporter.graph_data import GraphData, GraphDataReader, NodeInfoFactory


class KGNameSearcher(gensim.utils.SaveLoad):
    """
    search some node in KG by some keyword.
    """

    def __init__(self):
        self.full_name_2_id_set_map = {}
        self.id_2_full_name_set_map = {}
        self.keyword_2_id_set_map = {}
        self.id_2_keyword_set_map = {}

    def add_full_name_for_id(self, name, id):
        if not name:
            return
        if name not in self.full_name_2_id_set_map.keys():
            self.full_name_2_id_set_map[name] = set([])
        if id not in self.id_2_full_name_set_map.keys():
            self.id_2_full_name_set_map[id] = set([])

        self.full_name_2_id_set_map[name].add(id)
        self.id_2_full_name_set_map[id].add(name)

    def add_keyword_for_id(self, keyword, id):
        if not keyword:
            return
        if keyword not in self.keyword_2_id_set_map.keys():
            self.keyword_2_id_set_map[keyword] = set([])
        if id not in self.id_2_keyword_set_map.keys():
            self.id_2_keyword_set_map[id] = set([])

        self.keyword_2_id_set_map[keyword].add(id)

        self.id_2_keyword_set_map[id].add(keyword)

    def add_keyword_map_from_full_name(self, full_name, id):
        ## todo: change this to fix all

        full_name = full_name.split("(")[0]
        full_name = full_name.replace("-", " ").replace("(", " ").replace(")", " ").strip()

        self.add_keyword_for_id(full_name, id)

        name_words = full_name.split(" ")
        for word in name_words:
            self.add_keyword_for_id(word, id)
            self.add_keyword_for_id(word.lower(), id)

    @staticmethod
    def train_from_graph_data_file(graph_data_path, node_info_factory: NodeInfoFactory):
        """
        train the kg name searcher model from a graph data object
        :param node_info_factory: the nodeInfoFactory to create node from node json
        :param graph_data_path:
        :return:
        """
        graph_data: GraphData = GraphData.load(graph_data_path)

        searcher = KGNameSearcher()
        searcher.train(graph_data, node_info_factory)
        return searcher

    def train(self, graph_data: GraphData, node_info_factory: NodeInfoFactory):
        graph_data_reader = GraphDataReader(graph_data=graph_data, node_info_factory=node_info_factory)
        # todo: change the read all node from KG by iterate

        for id in graph_data.get_node_ids():
            node_info = graph_data_reader.get_node_info(id)
            name_list = node_info.get_all_names()
            for name in name_list:
                self.add_full_name_for_id(name, id)
                self.add_full_name_for_id(name.lower(), id)
                self.add_keyword_map_from_full_name(name, id)
                self.add_keyword_map_from_full_name(name.lower(), id)

    def search_by_full_name(self, full_name):
        if full_name in self.full_name_2_id_set_map.keys():
            return self.full_name_2_id_set_map[full_name]
        full_name = full_name.lower()
        if full_name in self.full_name_2_id_set_map.keys():
            return self.full_name_2_id_set_map[full_name]
        return set([])

    def search_by_keyword(self, word):
        if word in self.keyword_2_id_set_map.keys():
            return self.keyword_2_id_set_map[word]
        word = word.lower()

        if word in self.keyword_2_id_set_map.keys():
            return self.keyword_2_id_set_map[word]

        return set([])

    def get_full_names(self, id):
        if id not in self.id_2_full_name_set_map:
            return set([])
        else:
            return self.id_2_full_name_set_map[id]

    def clear(self):
        self.full_name_2_id_set_map = {}
        self.id_2_full_name_set_map = {}
        self.keyword_2_id_set_map = {}
        self.id_2_keyword_set_map = {}



import numpy as np


class VectorUtil:
    @staticmethod
    def get_weight_mean_vec(vector_list, weight_list):
        """
        get the average word2vec for list of str
        :param weight_list:
        :param vector_list:
        :return: np.array()
        """
        # todo: add a empty zero vectors result.
        # weight_sum = np.sum(weight_list)
        # normal_weight_list = []
        # for w in weight_list:
        #     normal_weight_list.append(w / weight_sum)
        # weight_list=normal_weight_list

        x = np.matrix(vector_list)
        avg_vector = np.average(x, axis=0, weights=weight_list)
        avg_vector = avg_vector.getA()[0]
        return avg_vector

    @staticmethod
    def compute_idf_weight_dict(total_num, number_dict):
        index_2_key_map = {}

        index = 0

        count_list = []
        for key, count in number_dict.items():
            index_2_key_map[index] = key
            count_list.append(count)
            index = index + 1

        a = np.array(count_list)
        ## smooth, in case the divide by zero error
        a = np.log((total_num + 1) / (a + 1))
        result = {}

        for index, w in enumerate(a):
            key = index_2_key_map[index]
            result[key] = w

        return result


"""
this class is used to get a weight for each relation and node.
"""



from unittest import TestCase

from sqlalchemy import Integer, Column, String
from sqlalchemy.ext.declarative import declarative_base

from sekg.mysql.accessor import MySQLAccessor
from sekg.mysql.factory import MysqlSessionFactory

Base = declarative_base()


class APIModel(Base):
    __tablename__ = 'api_test_model'
    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String(255), index=True, nullable=True)

    def __init__(self, name):
        self.name = name


class TestMySQLAccessor(TestCase):
    def test_delete_all(self):
        factory = MysqlSessionFactory("mysql_config.json")
        engine = factory.create_mysql_engine_by_server_id(1, database="testsekg")
        mysql_accessor = MySQLAccessor(engine=engine)
        mysql_accessor.create_orm_tables(SqlachemyORMBaseClass=Base)
        mysql_accessor.delete_all(APIModel)

    def test_create_table(self):
        factory = MysqlSessionFactory("mysql_config.json")
        engine = factory.create_mysql_engine_by_server_id(1, database="test", echo=True)
        mysql_accessor = MySQLAccessor(engine=engine)
        mysql_accessor.create_orm_tables(SqlachemyORMBaseClass=Base)
        mysql_accessor.drop_orm_tables(SqlachemyORMBaseClass=Base)

    def test_get_by_primary_key(self):
        factory = MysqlSessionFactory("mysql_config.json")
        engine = factory.create_mysql_engine_by_server_id(1, database="test", echo=True)
        mysql_accessor = MySQLAccessor(engine=engine)
        mysql_accessor.create_orm_tables(SqlachemyORMBaseClass=Base)

        api = mysql_accessor.get_by_primary_key(model_class=APIModel, primary_property=APIModel.id,
                                                primary_property_value=3)
        self.assertIsNone(api)
        # todo, add method for more



#!/usr/bin/python
# -*- coding: UTF-8 -*-
import pickle
from pathlib import Path

from sekg.ir.doc.wrapper import PreprocessMultiFieldDocumentCollection
from sekg.ir.models.base import DocumentSimModel
from sekg.util.vector_util import VectorUtil


# todo: change the model in here to save the model to local path
class CompoundSearchModel(DocumentSimModel):
    __sub_search_model_config__ = "submodel.config"
    __entity_collection_name__ = 'entity.collection'

    def __init__(self, name, model_dir_path, **config):

        """
        init the lsi model with
        :param model_dir_path:
        """
        super().__init__(name, model_dir_path, **config)
        self.sub_search_model_config = []
        self.preprocess_doc_collection = None
        self.model_list = []
        self.model_weight_list = []

        self.__init_sub_model_path__()

    def save(self, model_dir_path):
        super().save(model_dir_path)
        self.__init_sub_model_path__()
        with open(self.sub_search_model_config_path, 'wb') as out:
            out.write(pickle.dumps(self.sub_search_model_config))
        print("saving the sub_search_model_config complete")

        print("entity collection saving...")
        self.preprocess_doc_collection.save(self.entity_collection_path)
        print(
            "entity collection finish saving , save to %s, %r" % (
                self.entity_collection_path, self.preprocess_doc_collection))

    def __init_sub_model_path__(self):
        if self.model_dir_path is None:
            return
        Path(self.model_dir_path).mkdir(exist_ok=True, parents=True)
        # init the paths of models
        self.sub_search_model_config_path = str(Path(self.model_dir_path) / self.__sub_search_model_config__)
        self.entity_collection_path = str(Path(self.model_dir_path) / self.__entity_collection_name__)

    def init_model_as_submodel(self):
        self.init_model()

    def init_model(self):
        self.__init_sub_model_path__()

        with open(self.sub_search_model_config_path, 'rb') as aq:
            self.sub_search_model_config = pickle.loads(aq.read())
        self.init_all_sub_model()
        if self.preprocess_doc_collection is None:
            self.preprocess_doc_collection = PreprocessMultiFieldDocumentCollection.load(self.entity_collection_path)

    def init_all_sub_model(self):
        for model_dir_path, model_class, model_weight, load_model in self.sub_search_model_config:
            if not load_model:
                self.model_list.append(model_class.load_as_submodel(model_dir_path))
            else:
                model = model_class.load(model_dir_path)
                self.model_list.append(model)
                self.preprocess_doc_collection = model.preprocess_doc_collection
            self.model_weight_list.append(model_weight)

    def get_full_doc_score_vec(self, query):
        full_entity_score_vec = self.get_cache_score_vector(query)
        if full_entity_score_vec is not None:
            return full_entity_score_vec

        submodel_full_entity_score_vec_list = []
        for model in self.model_list:
            vector = model.get_full_doc_score_vec(query)
            submodel_full_entity_score_vec_list.append(vector)
        full_entity_score_vec = VectorUtil.get_weight_mean_vec(vector_list=submodel_full_entity_score_vec_list,
                                                               weight_list=self.model_weight_list)

        self.cache_entity_score_vector(query, full_entity_score_vec)
        return full_entity_score_vec

    def train_from_doc_collection_with_preprocessor(self, doc_collection, **config):
        """

        :param doc_collection:
        :param config: "search_model_class_list", each one is (model_path,model_class,model_weight)
        :return:
        """
        self.preprocess_doc_collection = doc_collection
        sub_search_model_config = []

        if "sub_search_model_config" in config.keys():
            sub_search_model_config = config["sub_search_model_config"]

        self.sub_search_model_config = sub_search_model_config
        if len(self.sub_search_model_config) == 0:
            raise Exception("search_model_class_config_list is []!!!")
        self.init_all_sub_model()



from nltk.corpus import wordnet as wn


class WordUtil:
    @staticmethod
    def couldBeVerb(word):
        """
        check if a word could be a verb
        :param word:
        :return:
        """
        pairs = [tuple(synset.name().split(".")[:2]) for synset in wn.synsets(word)]
        tag_count = {}
        for name, tag in pairs:
            tag_count[tag] = tag_count.get(tag, 0) + 1
        if len(tag_count) == 0:
            return False
        if "v" in tag_count:
            return True
        return False

    @staticmethod
    def couldBeNoun(word):
        """
        check if a verb could be noun
        :param word:
        :return:
        """
        pairs = [tuple(synset.name().split(".")[:2]) for synset in wn.synsets(word)]
        tag_count = {}
        for name, tag in pairs:
            tag_count[tag] = tag_count.get(tag, 0) + 1
        if len(tag_count) == 0:
            return False
        if "n" in tag_count:
            return True
        return False

    @staticmethod
    def couldBeADJ(word):
        """
        check if a word could be adj
        :param word:
        :return:
        """
        pairs = [tuple(synset.name().split(".")[:2]) for synset in wn.synsets(word)]
        tag_count = {}
        for name, tag in pairs:
            tag_count[tag] = tag_count.get(tag, 0) + 1
        if len(tag_count) == 0:
            return False
        if "a" in tag_count or "s" in tag_count:
            return True
        return False



class Preprocessor:
    """
    a basic text preprocessor
    """

    def extract_words_for_query(self, text):
        """
        extract the keyword set for a query text
        :param text:
        :return:
        """
        return set(self.clean(text))

    def clean(self, text):
        """
        return a list of token from the text
        :param text: the text need to preprocess
        :return: list of str
        """
        return text.split()


class SimplePreprocessor(Preprocessor):
    remove_list = {"\n", "\t", "\r", "/", "*", ".", ";", "@", "{", "}", "<p>", "(", ")", "#", "=", ":", "+", "-",
                   "!",
                   "[", "]", ",", ":", "<", ">", "|", "\\", "&", "'", "?", "<", ">"}

    def remove_special_char(self, str):
        """
        input;str
        :return: str
        """
        new_str = str
        for item in self.remove_list:
            new_str = new_str.replace(item, " ").replace("  ", " ")
        return new_str.strip()

    def clean(self, text):
        """
        return a list of token from the text, only remove special char and lower the case
        :param text: the text need to preprocess
        :return: list of str
        """
        text = self.remove_special_char(text)
        return text.lower().split()



from pathlib import Path
from unittest import TestCase

from sekg.graph.util.name_searcher import KGNameSearcher
from test.data.definition import ROOT_DIR


class TestAVGNode2VectorModel(TestCase):

    def test_load_name_searcher(self):
        data_dir = Path(ROOT_DIR)
        kg_name_searcher_path = str(data_dir / "JabRef-2.6.v3.namesearcher")
        kg_name_searcher: KGNameSearcher = KGNameSearcher.load(kg_name_searcher_path)
        result = kg_name_searcher.search_by_keyword("pdf")
        print(result)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


import re

from bs4 import BeautifulSoup
from nltk import WordNetLemmatizer
from nltk.corpus import wordnet as wn, stopwords

from sekg.text.extractor.domain_entity.nlp_util import SpacyNLPFactory
from sekg.util.code import CodeElementNameUtil


class EntityExtractor(object):
    """
    extract useful information from text, eg. HTML text, comment style text, normal text
    """

    def __init__(self):
        self.nlp = SpacyNLPFactory.create_spacy_nlp_for_domain_extractor()
        self.pattern = re.compile(r"NP_\w+ of NP_\w+")
        self.stopwords = stopwords.words('english')
        self.stopwords.append("-PRON-")
        self.stopwords = set(self.stopwords)
        self.lemmatizer = WordNetLemmatizer()

        self.code_patterns = [
            re.compile(r'^(?P<ELE>[a-zA-Z0-9_]*[a-z0-9][A-Z][a-z]+[a-zA-Z0-9_]*)(<.*>)?$'),
            re.compile(r'^(?P<ELE>[a-zA-Z0-9_\.<>]+)\([a-zA-Z0-9_\,.<>)]*?$'),
            re.compile(r'^(?P<ELE>[a-zA-Z]{2,}(\.[a-zA-Z0-9_]+)+)(<.*>)?$')
        ]

        self.camel_cache = {}
        self.CODE_NAME_UTIL = CodeElementNameUtil()

    def uncamelize(self, camel_case):
        if camel_case in self.camel_cache:
            return self.camel_cache[camel_case]
        sub = self.CODE_NAME_UTIL.uncamelize_by_stemming(camel_case)
        self.camel_cache[camel_case] = sub
        return sub

    def extract_from_sentence(self, sent):
        """
        extract concept from one sentence.
        :param sent:
        :return: a set of concepts.
        """
        code_elements = self.extract_code_element(sent)

        domain_terms = set()
        doc = self.nlp(sent)
        for chunk in doc.noun_chunks:
            # print("chunk: ", chunk.text)
            chunk = self.clean_chunk(chunk)
            # print("cleaned chunk:", chunk)
            if len(chunk) == 0:
                continue
            if len(chunk) == 1 and self.is_word_common(chunk.text):
                continue
            if chunk.text in code_elements:
                continue
            domain_terms.add(self.__chunk_lemmatize(chunk))
            domain_terms.update(self.extract_abbreviation_from_chunk(chunk))
            domain_terms.update(self.extract_NNPs_from_chunk(chunk))
        domain_terms.update(self.extract_np_of_np(doc))
        # print('sent: ' + sent)
        # print('result: ', result)
        domain_terms = self.__post_process(domain_terms)
        return domain_terms, code_elements

    def extract_code_element(self, sent):

        elements = set()
        for word in sent.split():
            word = word.lstrip("#(").rstrip(",;.!?")
            for pattern in self.code_patterns:
                search_rs = pattern.search(word)
                if search_rs is not None:
                    elements.add(search_rs.group("ELE"))
        return elements

    def extract_np_of_np(self, doc):
        result = set([])
        sentence_text = doc[:].lemma_
        for chunk in doc.noun_chunks:
            chunk_arr = []
            chunk = self.clean_chunk(chunk)
            if len(chunk) == 0:
                continue
            for token in chunk:
                chunk_arr.append(token.lemma_)
            chunk_lemma = " ".join(chunk_arr)
            # print("chunk_lemma", chunk_lemma)
            replacement_value = "NP_" + "_".join(chunk_arr)
            # print("replacement_value", replacement_value)
            sentence_text = sentence_text.replace(chunk_lemma, replacement_value)
        # print("sentence_text", sentence_text)
        matches = re.findall(self.pattern, sentence_text)
        if len(matches) > 0:
            # print('matched: ', matches)
            for m in matches:
                result.add(m.replace("NP_", "").replace("_", " "))
        return result

    def clean_chunk(self, chunk):
        """
        remove the stopwords, digits and pronouns at the start of the chunk.
        pass the result which contains invalid symbol.
        :param chunk:
        :return:
        """
        if chunk.text.lower() in self.stopwords:
            return []
        while len(chunk) > 1:
            start_token = chunk[0]
            if start_token.text.lower() in self.stopwords or start_token.text.isdigit() or start_token.tag_ == 'PRP':
                chunk = chunk[1:]
            else:
                break
        if len(chunk) == 1:
            start_token = chunk[0]
            if start_token.text.lower() in self.stopwords or start_token.text.isdigit() or start_token.tag_ == 'PRP':
                return []
        if not re.match(r'^[a-zA-Z0-9][a-zA-Z0-9\' -]*[a-zA-Z0-9]$', chunk.text):
            return []
        return chunk

    def is_word_common(self, word):
        """
        check if the word is common word.
        :param word:
        :return:
        """
        if word in self.stopwords:
            return True
        if re.match(r'[a-zA-Z]+[a-zA-Z]$', word):
            word = self.lemmatizer.lemmatize(word, pos='n')
            synset = wn.synsets(word)
            if len(synset) > 0:
                return True
            else:
                return False
        return False

    def extract_abbreviation_from_chunk(self, chunk):
        result = set([])
        for token in chunk:
            if re.match(r'[A-Z]{2,}[0-9]*$', token.text):
                result.add(token.text)
        return result

    def extract_NNPs_from_chunk(self, chunk):
        result = set([])
        p = 0
        while p < (len(chunk) - 1):
            if chunk[p].tag_.startswith('NNP'):
                for i in range(p + 1, len(chunk)):
                    if not chunk[i].tag_.startswith('NNP'):
                        t_w = chunk[p:i]
                        p = i
                        if len(t_w) > 1:
                            result.add(self.__chunk_lemmatize(t_w))
                        break
                    elif i == len(chunk) - 1:
                        t_w = chunk[p:]
                        p = i
                        if len(t_w) > 1:
                            result.add(self.__chunk_lemmatize(t_w))
                        break
            else:
                p = p + 1
        return result

    def __chunk_lemmatize(self, chunk):
        """
        lemmatize the last word of chunk.
        :param chunk:
        :return:
        """

        word = self.lemmatizer.lemmatize(chunk.text, pos='n')

        return word

    def __post_process(self, result):
        new_result = set([])
        for item in result:
            if len(item) == 1 or item.isdigit():
                continue
            new_result.add(item)
        return new_result

    def extract_from_comment(self, comment):
        """
        extract domain_terms, code_elements from comment text
        :param comment:
        :return:
        """
        comment = re.sub(r'\s+', ' ', comment.strip().strip("/*").strip())
        if len(comment) == 0:
            return set(), set()
        domain_terms, code_elements = self.extract_from_sentence(comment)
        return domain_terms, code_elements

    def extract_from_html(self, html):
        terms = set()
        soup = BeautifulSoup(html, "lxml")
        tts = {tt.get_text() for tt in soup.findAll("tt")}
        terms.update({tt for tt in tts if len(tt.split()) <= 3})
        sent = soup.get_text()
        sent = re.sub(r'\s+', ' ', sent.strip().strip("/*").strip())
        domain_terms, code_elements = self.extract_from_sentence(sent)
        for term in domain_terms:
            terms.add(term)
        return terms, code_elements



# -*- coding: utf-8 -*-s

from __future__ import absolute_import

import re

from sqlalchemy import event, literal, Index
from sqlalchemy.schema import DDL
from sqlalchemy.orm.mapper import Mapper
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql.expression import ClauseElement
from . import modes as FullTextMode


MYSQL = "mysql"
MYSQL_BUILD_INDEX_QUERY = u"""ALTER TABLE {0.__tablename__} ADD FULLTEXT ({1})"""
MYSQL_MATCH_AGAINST = u"""
                      MATCH ({0})
                      AGAINST ({1} {2})
                      """

def escape_quote(string):
    return re.sub(r"[\"\']+", "", string)

class FullTextSearch(ClauseElement):
    """
    Search FullText
    :param against: the search query
    :param table: the table needs to be query

    FullText support with in query, i.e.
         from sqlalchemy_fulltext import FullTextSearch
         session.query(Foo).filter(FullTextSearch('Spam', Foo))
    """
    def __init__(self, against, model, mode=FullTextMode.DEFAULT):
        self.model = model
        self.against = literal(against)
        self.mode = mode


def get_table_name(element):
    if hasattr(element.model, "__table__"):
        return "`" + element.model.__table__.fullname + "`."
    return ""


@compiles(FullTextSearch, MYSQL)
def __mysql_fulltext_search(element, compiler, **kw):
    assert issubclass(element.model, FullText), "{0} not FullTextable".format(element.model)
    return MYSQL_MATCH_AGAINST.format(
        ", ".join([get_table_name(element) + column for column in element.model.__fulltext_columns__]),
        compiler.process(element.against),
        element.mode)


class FullText(object):
    """
    FullText Minxin object for SQLAlchemy
    
         from sqlalchemy_fulltext import FullText
         class Foo(FullText, Base):
             __fulltext_columns__ = ('spam', 'ham')
             ...

    fulltext search spam and ham now
    """
    
    __fulltext_columns__ = tuple()

    @classmethod
    def build_fulltext(cls, table):
        """
        build up fulltext index after table is created
        """
        if FullText not in cls.__bases__:
            return
        if not cls.__fulltext_after_create__:
            return
        assert cls.__fulltext_columns__, "Model:{0.__name__} No FullText columns defined".format(cls)

        event.listen(table,
                     'after_create',
                     DDL(MYSQL_BUILD_INDEX_QUERY.format(cls,
                         ", ".join((escape_quote(c)
                                    for c in cls.__fulltext_columns__)))
                         )
                     )
    """
    TODO: black magic in the future
    @classmethod
    @declared_attr
    def __contains__(*arg):
        return True
    """

    __fulltext_after_create__ = True


class FullTextForMigration(FullText):
    __fulltext_after_create__ = False

    @classmethod
    def index_fulltext(cls):
        """
        call like Index('idx_<__tablename__>_fulltext', *__fulltext_columns__, mysql_prefix='FULLTEXT')
        """
        assert cls.__tablename__, "Model:{0.__name__} No table name defined".format(cls)
        assert cls.__fulltext_columns__, "Model:{0.__name__} No FullText columns defined".format(cls)
        Index('idx_%s_fulltext' % cls.__tablename__,
                *(getattr(cls, c) for c in cls.__fulltext_columns__),
                  mysql_prefix='FULLTEXT')


def __build_fulltext_index(mapper, class_):    
    if issubclass(class_, FullText):
        class_.build_fulltext(mapper.mapped_table)


event.listen(Mapper, 'instrument_class', __build_fulltext_index)



import traceback

from sqlalchemy import text
from sqlalchemy.orm import sessionmaker

from .factory import MysqlSessionFactory


class MySQLAccessor:
    """
        This class wrap some query operation for mysql table with wrapping session, first you init this MySQLAccessor with a session.
        The session object is obtained by MysqlSessionFactory, it is from sqlalchemy library, connecting to one database
        in one mysql server, than you can use this accessor to do some query.
        You can extends this accessor to add more custom function.
    """

    def __init__(self, engine, autocommit=False, echo=True):
        self.engine = engine
        self.session = MysqlSessionFactory.create_mysql_session_from_engine(engine=engine, autocommit=autocommit,
                                                                            echo=echo)

    @staticmethod
    def create_mysql_session_from_engine(engine, autocommit, echo):
        """
        create the session to one specify database from one server
        :param engine: the engine use to create engine
        :param echo: if true, all the sql executed will print to console
        :param autocommit: if True, all sql will be commit automately. If False,all write sql statement will be executed after session.commit() call.

        :return: the session object in sqlalchemy, None if create fail
        """
        if engine is None:
            print("engine create fail for create session")
            return None
        Session = sessionmaker(bind=engine, autocommit=autocommit)
        session = Session()

        if echo:
            print("create new session by %r" % autocommit)
        return session

    def is_connect(self):
        """
        check if the session valid
        :return: True, valid; False, not valid
        """
        if self.session is None:
            return False
        else:
            return True

    def delete_all(self, model_class):
        """
        delete all item in one table
        :param model_class: the model class extend Base, "Base = declarative_base()", it has a __tablename__ property
        to find the table
        :return:
        """
        try:
            self.session.query(model_class).delete()
            self.session.commit()
        except Exception as error:
            traceback.print_exc()
            return

    def create_orm_tables(self, SqlachemyORMBaseClass):
        """
        create all table, must given a SqlachemyORMBaseClass
        :param SqlachemyORMBaseClass: SqlachemyORMBaseClass that every table model extended
        :return:
        """
        # create the table
        SqlachemyORMBaseClass.metadata.create_all(bind=self.engine)

    def drop_orm_tables(self, SqlachemyORMBaseClass):
        """
                create all table, must given a SqlachemyORMBaseClass
                :param SqlachemyORMBaseClass: SqlachemyORMBaseClass that every table model extended
                :return:
        """

        # delete all table
        SqlachemyORMBaseClass.metadata.drop_all(bind=self.engine)

    def create_table_by_metadata(self, metadata):
        """
        create a table by a metadata
        example: create table by metadata
        metadata = MetaData(engine)

        user = Table('user', metadata,
        Column('id', Integer, primary_key = True),
        Column('name', String(20)),
        Column('fullname', String(40)))
        address = Table('address', metadata,
            Column('id', Integer, primary_key = True),
            Column('user_id', None, ForeignKey('user.id')),
            Column('email', String(60), nullable = False),
        )
        metadata.create_all(engine)

        :param metadata:
        :return:
        """
        metadata.create_all(engine=self.engine)

    def get_by_primary_key(self, model_class, primary_property, primary_property_value):
        """
        get one Model class object by query the Mysql, the primary property value must be the primary_property_value.

        example: model_class=APIEntity,primary_property="id",primary_property_value="3", get the apiEntity where "id"=3
               :param model_class: the model class extend Base, "Base = declarative_base()", it has a __tablename__ property
               to find the table
               :param primary_property: the primary_property in model_class to find a unique object. etc. APIEntity.id
                :param primary_property_value: the primary_property_value that primary_property is

               :return:
        """

        try:
            return self.session.query(model_class).filter(primary_property == primary_property_value).first()
        except Exception:
            traceback.print_exc()
            return None

    def add_index(self, index_name, table_name, column_name):
        """
        add index to a table in one column
        :param index_name: the index name
        :param table_name: the table name
        :param column_name: the column name
        :return:
        """
        # todo: need test
        try:
            conn = self.engine.connect()
            text_sql = 'alter table {table_name} add index {index_name}(column_name)'.format(
                table_name=table_name,
                index_name=index_name,
                column_name=column_name)

            s = text(text_sql)

            conn.execute(s)
            conn.close()
        except:
            traceback.print_exc()

    def add_multi_index(self, index_name, table_name, *column_name_list):
        """
        add index to a table in one column
        :param index_name: the index name
        :param table_name: the table name
        :param column_name_list: the column name list
        :return:
        """
        # todo: need test
        try:
            conn = self.engine.connect()
            text_sql = 'alter table {table_name} add index {index_name}(column_name)'.format(
                table_name=table_name,
                index_name=index_name,
                column_name=",".join(column_name_list))

            s = text(text_sql)

            conn.execute(s)
            conn.close()
        except:
            traceback.print_exc()

    def delete_multi_index(self, index_name, table_name, *column_name_list):
        """
        delete index to a table in one column
        :param index_name:  the index name
        :param table_name: the table name
        :param column_name_list: the column name list
        :return:
        """
        # todo: need test

        try:
            conn = self.engine.connect()
            text_sql = 'alter table {table_name} drop index {index_name}(column_name)'.format(
                table_name=table_name,
                index_name=index_name,
                column_name=",".join(column_name_list))

            s = text(text_sql)

            conn.execute(s)
            conn.close()
        except:
            traceback.print_exc()

    def query_all(self, model_class):
        """
        query all item in one table
        :param model_class: the model class extend Base, "Base = declarative_base()", it has a __tablename__ property
        to find the table
        :return: the all result
        """
        try:
            self.session.query(model_class).all()
        except Exception as error:
            traceback.print_exc()
            return []

    def create_without_duplicate(self, autocommit=True):
        # todo: fix this method
        pass



import abc
import os
import shutil
from pathlib import Path

import numpy as np

from sekg.ir.doc.wrapper import PreprocessMultiFieldDocumentCollection
from sekg.ir.preprocessor.base import Preprocessor


class DocRetrievalResult:
    """
    this class is wrapper for document retrieval result. Containing doc_id, doc_name,score.
    """

    def __init__(self, doc_id, ranking, score, doc_name, **extra_info):
        """
        create a wrapper for document search result.
        :param doc_id: the doc id.
        :param ranking: the ranking of the doc.
        :param score: the score of this document
        :param doc_name: the name of this document, could be None and duplicate
        :param extra_info: some extra info with this document
        """
        self.doc_id = doc_id
        self.doc_name = doc_name
        self.score = score
        self.ranking = ranking
        self.extra_info = extra_info

    def __repr__(self):
        return "<DocumentRetrieval result: rank=%d id=%d score=%r name=%s>" % (
            self.ranking, self.doc_id, self.score, self.doc_name,)


# todo: the index, doc id, score index is confuse, must be fixed. could be all use the PreprocessDocCollection as a index defined.
# todo: make the documentSimModel could train from normal document collection
class DocumentSimModel(metaclass=abc.ABCMeta):
    """
    the basic class for Model that computes the similarity score between query and documents.
    This is responsible for training the documentSim model and save and load the models, cache the model.
    And it could get the document similar to query by sim score.
    the sim score is between [0,1].
    Ps. each model is save under a directory. And all model file related to this model is put in this dir.
    """

    def __init__(self, name, model_dir_path, **config):
        """
        init the graph search model.
        :param name: the name of the model, just name the model instance, not other meaning.
        :param model_dir_path: the mode save dir path
        :param config: the config of the document similar model, eg. the weight of sub models
        """
        self.name = name
        self.model_dir_path = model_dir_path
        self.config = config
        self.preprocess_doc_collection = None
        # use to cache the score vector, score vector is a vector v
        # v=[0.5,2.0,3.0], v[0] means that the document 'd' ,
        # whose index is 0 in document collection, its score with query is 0.5.
        self.query_2_score_vector_cache = {}
        # sorted_index_scores is that score_vector sort by score from hign to low and keep the document index.
        self.query_2_sorted_index_scores_cache = {}
        self.preprocessor = None

    @abc.abstractmethod
    def init_model(self):
        """
        init the model, need to implement by each model
        :return:
        """
        print("start init the model")

    @abc.abstractmethod
    def init_model_as_submodel(self):
        """
        init the model as submodel.
        only load some some neccessary data for compute get_full_entity_score_vec
        :return:
        """
        print("start init model as sub_model ")

    def search(self, query, top_num=10, valid_doc_id_set=None):
        """
        search some document by query, return top score document by similar. and the document id must in the given id set.
        if valid_doc_id_set is not given. any document could be retrieved.
        :param valid_doc_id_set: the doc id set. the id of the returned document, its doc_id must in this id set.
        if None, any document could be returned.
        :param top_num: the number of return result, if the top_num =0, return all result in valid_doc_id_set
        :param query: a str, stand for the query
        :return: list of DocRetrievalResult.
        """
        if valid_doc_id_set is None:
            valid_doc_id_set = {}

        sorted_index_scores = self.get_sorted_index_doc_scores(query)
        retrieval_results = self.filter_retrieve_document(sorted_index_scores,
                                                          top_n=top_num,
                                                          valid_doc_id_set=valid_doc_id_set)

        return retrieval_results

    @abc.abstractmethod
    def get_full_doc_score_vec(self, query):
        """
        score vector is a vector v=[0.5,2.0,3.0], v[0] means that the document 'd',
         whose doc index(or doc_position) is 0 in document collection, its score with query is 0.5.
        :param query: a str stands for the query.
        :return: get all document similar score with query as a numpy vector. The score for each doc is range from [0,1]
        """
        pass

    @abc.abstractmethod
    def train_from_doc_collection_with_preprocessor(self, doc_collection: PreprocessMultiFieldDocumentCollection,
                                                    **config):
        """
        train the doc collection with preprocess document collection
        :param doc_collection: doc_collection with preprocessor
        :return: a model instance
        """

        print("train_from_doc_collection_with_preprocessor config=%r" % config)
        self.preprocess_doc_collection = doc_collection
        self.config = config

    @classmethod
    def model_type(cls):
        """
        get the model type name
        :return: the model type name
        """
        return cls.__name__

    @classmethod
    def train(cls, model_dir_path, doc_collection: PreprocessMultiFieldDocumentCollection, **config):
        model = cls(cls.__name__, model_dir_path)
        model.train_from_doc_collection_with_preprocessor(doc_collection, **config)
        model.save(model_dir_path)
        return model

    @classmethod
    def load(cls, model_dir_path, **config):
        """
        load the model from a dir.
        :param model_dir_path: the dir saved a model
        :return: a model instance
        """
        model = cls.__create_from_a_exist_model_dir(config, model_dir_path)
        model.init_model()
        return model

    @classmethod
    def load_as_submodel(cls, model_dir_path, **config):
        """
        load the model from a dir.
        :param model_dir_path: the dir saved a model
        :return: a model instance
        """
        model = cls.__create_from_a_exist_model_dir(config, model_dir_path)
        model.init_model_as_submodel()
        return model

    @classmethod
    def __create_from_a_exist_model_dir(cls, config, model_dir_path):
        if model_dir_path is None:
            raise Exception("the model dir is None!!!")
        if not Path(model_dir_path).exists():
            raise Exception("the model dir %s is not exist" % model_dir_path)
        if not Path(model_dir_path).is_dir():
            raise Exception("the model dir %s is not not dir!!!" % model_dir_path)
        print("load the model from %s config=%r " % (model_dir_path, config))
        model = cls(cls.__name__, model_dir_path, **config)
        return model

    def save(self, model_dir_path):
        """
        save the model to a dir.
        :param model_dir_path: the dir to save a model
        :return:
        """
        if model_dir_path is None:
            raise Exception("the model dir is None!!!")

        if not Path(model_dir_path).exists():
            print("the model dir %s is not exist, creating" % model_dir_path)
            Path(model_dir_path).mkdir(parents=True, exist_ok=True)
        if not Path(model_dir_path).is_dir():
            raise Exception("the model dir %s is not not dir!!!" % model_dir_path)

        if len(os.listdir(model_dir_path)) != 0:
            print("the model dir %s is not empty before saving" % model_dir_path)
            shutil.rmtree(model_dir_path)
        print("save the model path")
        self.model_dir_path = model_dir_path

    def doc_index2doc_id(self, index):
        """
        parse the index to doc id
        :param index: the index of the document object in similarity_vec int value
        :return: the doc id , integer, not the str
        """

        return self.preprocess_doc_collection.doc_index_to_doc_id(index)

    def doc_id2doc_index(self, doc_id):
        return self.preprocess_doc_collection.doc_id_to_doc_index(doc_id)

    def similarity_by_doc_id(self, doc_id):
        """

        :param doc_id:
        :return:
        """
        # todo: implement in the base class or set as a abstract class
        pass

    def doc_id2doc(self, doc_id):
        """
        parse the entity_id to entity doc object (MultiFieldDocument)
        :return: the entity doc object(MultiFieldDocument)
        """
        return self.preprocess_doc_collection.get_preprocess_doc_by_id(doc_id)

    def doc_index2doc(self, index):
        """
        get the  MultiFieldDocument obj with the index
        :param index: the index of the MultiFieldDocument object in all PDocumentCollection
        :return: MultiFieldDocument
        """
        return self.preprocess_doc_collection.get_preprocess_doc_by_index(index)

    def similarity_4_doc_id_pair(self, doc_id1, doc_id2):
        # todo: implement in the base class or set as a abstract class
        pass

    def cache_entity_score_vector(self, query, index_scores):
        self.query_2_score_vector_cache[query] = index_scores

    def get_cache_score_vector(self, query):
        if query in self.query_2_score_vector_cache.keys():
            return self.query_2_score_vector_cache[query]
        return None

    def cache_sorted_index_scores(self, query, sorted_index_scores):
        self.query_2_sorted_index_scores_cache[query] = sorted_index_scores

    def get_cache_sorted_index_scores(self, query):
        if query in self.query_2_sorted_index_scores_cache.keys():
            return self.query_2_sorted_index_scores_cache[query]
        return None

    def get_doc_id_with_score(self, query):
        """
        get list of (doc_id,score) tuple.
        :param query:
        :return: a list
        """
        score_vector = self.get_full_doc_score_vec(query)
        doc_id_scores = [(self.doc_index2doc_id(index), score) for index, score in enumerate(score_vector)]
        return doc_id_scores

    def get_sorted_index_doc_scores(self, query):
        sorted_index_scores = self.get_cache_sorted_index_scores(query)
        if sorted_index_scores is None:
            score_vector = self.get_full_doc_score_vec(query)
            sort_index = np.argsort(-score_vector)
            score_vector = score_vector[sort_index]
            sorted_index_scores = np.array((sort_index, score_vector)).T
            # index_scores = [(index, score) for index, score in enumerate(score_vector)]
            # sorted_index_scores = sorted(index_scores, key=lambda item: item[1], reverse=True)
            self.cache_sorted_index_scores(query, sorted_index_scores)

        return sorted_index_scores

    def filter_retrieve_document(self, sorted_index_scores, top_n, valid_doc_id_set=None):
        """
        filter the retrieve result.
        First, remove all doc not in valid doc id set(if valid doc id set is given, otherwise the all doc are keep).
        Second. get the specific num of document with top score
        :param sorted_index_scores: list of (index:int,score:double), it is sorted by the score from from high to low.
        :param top_n:if top n=0, the num of result is not limit.
        :param valid_doc_id_set: the valid doc id set.
        :return: a list of DocRetrievalResult, sorted by score
        """
        if not valid_doc_id_set:
            retrieval_results = self.__get_retrieval_result(sorted_index_scores, top_n)
        else:
            retrieval_results = self.__get_valid_retrieval_result(sorted_index_scores, top_n, valid_doc_id_set)

        return retrieval_results

    @staticmethod
    def __create_entity_retrieval_result(document_id, entity_document, ranking, score):
        single_result = DocRetrievalResult(doc_id=document_id, ranking=ranking, score=score,
                                           doc_name=entity_document.get_name(), entity_document=entity_document)
        return single_result

    def __get_valid_retrieval_result(self, sorted_index_scores, top_n, valid_doc_id_set):
        valid_doc_index_set = self.preprocess_doc_collection.doc_id_set_2_doc_index_set(valid_doc_id_set)
        if top_n == 0:
            # return all valid result.
            retrieval_results = []
            ranking = 1
            for (doc_index, score) in sorted_index_scores:
                entity_document = self.doc_index2doc(doc_index)
                if entity_document is None:
                    continue
                if len(valid_doc_index_set) > 0 and doc_index not in valid_doc_index_set:
                    continue
                document_id = entity_document.get_document_id()
                if document_id not in valid_doc_id_set:
                    continue

                single_result = self.__create_entity_retrieval_result(document_id, entity_document, ranking, score)

                retrieval_results.append(single_result)
                ranking += 1

            return retrieval_results
        else:
            retrieval_results = []
            ranking = 1
            for (doc_index, score) in sorted_index_scores:
                entity_document = self.doc_index2doc(doc_index)
                if entity_document is None:
                    continue
                if len(valid_doc_index_set) > 0 and doc_index not in valid_doc_index_set:
                    continue
                document_id = entity_document.get_document_id()
                if document_id not in valid_doc_id_set:
                    continue
                single_result = self.__create_entity_retrieval_result(document_id, entity_document, ranking, score)

                retrieval_results.append(single_result)
                ranking += 1

                if ranking > top_n:
                    break
            return retrieval_results

    def __get_retrieval_result(self, sorted_index_scores, top_n):
        if top_n == 0:
            retrieval_results = []
            ranking = 1
            for (doc_index, score) in sorted_index_scores:
                entity_document = self.doc_index2doc(doc_index)
                if entity_document is None:
                    continue
                document_id = entity_document.get_document_id()

                single_result = self.__create_entity_retrieval_result(document_id, entity_document, ranking, score)
                retrieval_results.append(single_result)
                ranking += 1

            return retrieval_results
        else:
            retrieval_results = []
            ranking = 1
            for (doc_index, score) in sorted_index_scores[:top_n]:
                entity_document = self.doc_index2doc(doc_index)
                if entity_document is None:
                    continue
                document_id = entity_document.get_document_id()
                single_result = self.__create_entity_retrieval_result(document_id, entity_document, ranking, score)
                retrieval_results.append(single_result)
                ranking += 1
            return retrieval_results

    def set_preprocess_doc_collection(self, preprocess_doc_collection: PreprocessMultiFieldDocumentCollection):
        """
        set the document collection for DocumentSimModel
        :param preprocess_doc_collection:
        :return:
        """
        self.preprocess_doc_collection = preprocess_doc_collection

    def get_preprocess_doc_collection(self):
        return self.preprocess_doc_collection

    def set_preprocessor(self, preprocessor: Preprocessor):
        """
        set the preprocessor used to pre-process the query, eg. remove stop words and special char.
        This preprocessor depends on different
        :param preprocessor:
        :return:
        """
        self.preprocessor = preprocessor

    def __init_document_collection(self, preprocess_doc_collection: PreprocessMultiFieldDocumentCollection):
        ##todo rename this method
        self.set_preprocess_doc_collection(preprocess_doc_collection)
        self.set_preprocessor(preprocess_doc_collection.get_preprocessor())
